"""
Tagfu - An Entity Tagging System
Copyright (c) 2006, Alexander Botero-Lowry, Sham Chukoury, Michael Richardson
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:
  * Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, 
    this list of conditions and the following disclaimer in the documentation 
    and/or other materials provided with the distribution.
  * Neither the name of the authors explicitly named above, the name of the 
    Tagfu Team, nor the names of its contributors may be used to endorse or 
    promote products derived from this software without specific prior written 
    permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
import os

from pysqlite2 import dbapi2 as sqlite
from shutil import move

import tagfu_attrs

DB_VERSION=1

class Tags:
	def __init__(self, con, cur):
		self.con = con
		self.cur = cur

		self.con.create_function('change_tag', -1, self._tag_change)

		self.tags = []
		self._init_tags()

	def _tag_change(self):
		# FIXME: Should handle atomic changes instead of updating everything
		self.tags = []
		for row in self.con.execute("select * from Tag order by id;"):
			self.tags.append((row[0], row[1]))

		# TODO: Investigate whether con.commit is necessary as I have received
		# reports of it throwing an exception.
		try:
			self.con.commit()
		except:
			pass

	def _init_tags(self):
		self.tags = []
		for row in self.con.execute("select * from Tag order by id;"):
			self.tags.append((row[0], row[1]))

	def __repr__(self):
		""" Shows a list of tags """
		names = [ x[1] for x in self.tags ]
		return repr(names)

	def __len__(self):
		""" Gives the number of tags in the database """
		self.cur.execute('select count(distinct id) from Tag;')
		tag_count = self.cur.fetchone()
		return tag_count[0]

	def __contains__(self, item):
		""" Returns True is a tag by the name or id of item exists """
		ids = [ x[0] for x in self.tags ]
		names = [ x[1] for x in self.tags ]
		if item in ids or item in names:
			return True
		else:
			return False

	def __getitem__(self, item):
		""" 
			Returns the name of tag at id item, or the id of tag with name item 
		""" 
		if isinstance(item, basestring):
			for a in self.tags:
				if a[1] == item:
					return a[0]
		else:
			return self.tags[item]

	def __delitem__(self, item):
		""" Deletes a tag and all references to it """
		id = self.dbase_id(item)
		self.cur.execute('DELETE FROM Tag where id=?', (id,))
		self.cur.execute('DELETE FROM EntityTags where tag_id=?', (id,))
		self.con.commit()

	def __iter__(self):
		""" Returns an iterator for all tags """
		return iter(self.tags)
	
	def _new_tag(self, name):
		try:
			self.cur.execute('insert into Tag(name) values (?);', (name,))
			self.con.commit()
		except sqlite.IntegrityError, err:
			if str(err) == 'column name is not unique':
				raise ValueError("Creation of new tag failed: Tag already exists")
			else:
				raise ValueError("Creation of new tag failed: %s" % err)
		else:
			self.cur.execute("select max(id) from Tag")
			return self.cur.fetchone()[0]

	def append(self, name):
		""" Add a new tag to the list """
		if isinstance(name, basestring):
			return self._new_tag(name)
		else:
			raise TypeError('Tag name must be a string')

	def index(self, name):
		""" Get the index for tag with name name """
		for a in self.tags:
			if a[1] == name:
				return self.tags.index(a)
		
	def dbase_id(self, name):
		""" get the database id for tag by name or index """
		if isinstance(name, basestring):
			for a in self.tags:
				if a[1] == name:
					return a[0]
		else:
			return self.tags[name][0]

class Entities:
	def __init__(self, con, cur, alltags):
		self.con = con
		self.cur = cur
		self._ent_tags = 0
		self.alltags = alltags
		self.entities = []

	def _check_ents(self):
		self.cur.execute('select count(id) from Entity;')
		ent_count = self.cur.fetchone()
		if ent_count == self._ent_tags:
			return True
		else:
			self._ent_tags = ent_count
			return False

	def _update_ents(self):
		if not self._check_ents():
			self.entities = []

			for row in self.con.execute('select id, url from Entity;'):
				self.entities.append(Entity((self.con, self.cur), row, self.alltags))

	def __repr__(self):
		""" Return a representation of the pool of entities """
		self._update_ents()
		return repr(self.entities)

	def __len__(self):
		""" Return the number of entitites """
		self.cur.execute('select count(distinct id) from Entity;')
		ent_count = self.cur.fetchone()
		return ent_count[0]

	def __getitem__(self, item):
		""" Return an Entity object for the requested entity """
		self._update_ents()
		if isinstance(item, int):
			for a in self.entities:
				if a['_id'] == item:
					return a
			raise IndexError('database id not in entities list')
		elif isinstance(item, basestring):
			for a in self.entities:
				if a['_url'] == item:
					return a
			raise KeyError('url not in entities list')
		elif isinstance(item, Entity):
			for a in self.entities:
				if a == item:
					return a
			raise KeyError('Entity not in entities list')
		raise TypeError('object must be a database id, a string, or an Entity object')
			
	def __delitem__(self, item):
		""" Delete an entity and all associated tags """
		self._update_ents()
		if isinstance(item, int):
			id = self[item]['_id']	
		else:
			id = self.dbase_id(item)
		
		if self[item]["_url"].startswith("file://") and os.path.exists(self[item]["_url"][7:]):
			tagfu_attrs.delete_tagfu_id(self[item]["_url"][7:])
		self.cur.execute('delete from EntityTags where entity_id=?;', (id,))
		self.cur.execute('delete from Entity where id=?;', (id,))
		self.con.commit()
			
	def __iter__(self):
		""" Return an iterator of all entities """
		self._update_ents()
		return iter(self.entities)
	
	def __contains__(self, item):
		""" Tests if the requested entity is in the list of entitites """
		try:
			self[item]
			return True
		except:
			return False
			
	
	def _new_entity(self, url):
		# check URL for a tagfu_id and if it has one, update _url
		if url.startswith("file://") and os.path.exists(url[7:]):
			tagfu_id = tagfu_attrs.get_tagfu_id(url[7:])
			if tagfu_id:
				self.entities[tagfu_id]['_url'] = url
				return
		try:
			self.cur.execute('insert into Entity(url) values (?);', (url,))
			self.con.commit()
		except sqlite.IntegrityError, err:
			if str(err) == 'column name is not unique':
				raise ValueError("Creation of new entity failed: Entity already exists")
			else:
				raise ValueError("Creation of new entity failed: %s" % err)
		else:
			self.cur.execute("select max(id) from Entity")
			id = self.cur.fetchone()[0]
			if url.startswith('file://') and os.path.exists(url[7:]):
				tagfu_attrs.set_tagfu_id(url[7:], id)
			return id
	
	def append(self, name):
		""" Add a new entity """
		if isinstance(name, basestring):
			return self._new_entity(name)
		else:
			raise TypeError('Entity name must be a string')

	def index(self, item):
		""" Return the index of an entity """
		if isinstance(item, basestring):
			for a in self.entities:
				if a['_url'] == item:
					return a['_id']
		elif isinstance(item, Entity):
			if item in self.entities:
				return self.entities[item]['_id']
		raise ValueError('entity not in entities list')

	def dbase_id(self, name):
		""" Return the database id of an entity """
		if isinstance(name, basestring):
			for a in self.entities:
				if a['_url'] == name:
					return a['_id']
		else:
			return self[name]['_id']

class Entity:
	def __init__(self, dbase, private_tags, taglist):
		self.con, self.cur = dbase
		self.tags = {}
		self.id, self.url = private_tags
		self.taglist = taglist
		self._update_ent_tags()

	def _check_ent_tags(self):
		ent_tags = self.__len__()
		if ent_tags == len(self.tags):
			return True
		else:
			return False

	def _update_ent_tags(self):
		self.tags = {}
		self.tags['_id'] = self.id
		self.tags['_url'] = self.url
		for row in self.con.execute('select t.name, et.properties from EntityTags et left join Entity e on e.id == et.entity_id left join Tag t on t.id = et.tag_id where e.id=?;', (self.id,)):
			self.tags[row[0]] = row[1]

	def __repr__(self):
		""" Return a string representation of the entity and tags """
		self._update_ent_tags()
		return repr(self.tags)

	def __getitem__(self, item):
		""" Return a tag value """
		self._update_ent_tags()
		return self.tags[item]

	def __delitem__(self, key):
		""" Delete an associated tag """
		self.cur.execute('DELETE FROM EntityTags where entity_id=? and tag_id=?', (self.id, self.taglist.dbase_id(key)))
		self.con.commit()

	def __setitem__(self, key, value):
		""" Set the value of an associated tag """
		self._update_ent_tags()
		if key == "_url":
			# Remove old tagfu_id if there is one, and add a new one to new url
			if value.startswith('file://') and os.path.exists(value[7:]):
				if not tagfu_attrs.get_tagfu_id(value[7:]):
					tagfu_attrs.set_tagfu_id(value[7:], self.id)
				if self.url.startswith('file://') and os.path.exists(self.url[7:]):
					tagfu_attrs.delete_tagfu_id(self.url[7:])
			self.cur.execute("UPDATE Entity set url=? where id=?", (value, self.id))
			self.con.commit()
		elif self.tags.has_key(key):
			self.cur.execute('UPDATE EntityTags set properties=? where entity_id=? and tag_id=?;', (value, self.id, self.taglist.dbase_id(key)))
			self.con.commit()
		else:
			if not key in self.taglist:
				self.taglist.append(key)

			if value:
				self.cur.execute('INSERT INTO EntityTags (entity_id, tag_id, properties) values	(?, ?, ?);', (self.id, self.taglist.dbase_id(key), value))
				self.con.commit()
			else:
				self.cur.execute('INSERT INTO EntityTags (entity_id, tag_id) values	(?, ?);', (self.id, self.taglist.dbase_id(key)))
				self.con.commit()
	
	def __contains__(self, tag):
		""" Test whether an entity has a tag associated to it """
		if tag in self.tags:
			return True
		else:
			return False

	def __iter__(self):
		""" Return an iterator of associated tags """
		self._update_ent_tags()
		return iter(self.tags)

	def __len__(self):
		""" Return a count of associated tags """
		self.cur.execute('select count(*) from EntityTags where entity_id=?;', (self.id,))
		ent_count = self.cur.fetchone()
		return ent_count[0]+2

	def iterkeys(self):
		""" Return an iterator of the tag names associated to the entity """
		self._update_ent_tags()
		return self.tags.iterkeys()
		
	def iteritems(self):
		""" Return an iterator of the tag items associated to the entity """
		self._update_ent_tags()
		return self.tags.iteritems()

	def itervalues(self):
		""" Return an iterator of the tag values associated to the entitty """
		self._update_ent_tags()
		return self.tags.iterkeys()

	def has_key(self, key):
		""" Test whether a tag is associated to the entity """
		self._update_ent_tags()
		if key in self:
			return True
		else:
			return False

class Tagfu:
	def __init__(self, dbpath=None):
		""" Create an instance of TagFu used for accessing a particular TagFu database instance """
		need_tables = False
		if dbpath:
			tagfu_path = dbpath
		else:
			tagfu_path = os.getenv("TAGFU_DB", os.path.join(os.path.expanduser('~'), '.tagfu', 'tagfu.db'))

		if not os.path.exists(os.path.dirname(tagfu_path)):
			os.makedirs(os.path.dirname(tagfu_path))
		if not os.path.exists(tagfu_path):
			need_tables = True

		self.con = sqlite.connect(tagfu_path)
		self.cur = self.con.cursor()
		con = self.con
		cur = self.cur

		cur.execute('PRAGMA user_version;')
		version = cur.fetchone()[0]

		if not need_tables and str(version) != str(DB_VERSION):
			con.close()
			move(tagfu_path, tagfu_path+'.old')
			need_tables = True
			self.con = sqlite.connect(tagfu_path)
			self.cur = self.con.cursor()
			con = self.con
			cur = self.cur

		if need_tables == True:
			cur.execute('PRAGMA user_version = "'+str(DB_VERSION)+'";')

			cur.execute("CREATE TABLE Entity (id integer primary key AUTOINCREMENT, url UNIQUE);")
			cur.execute("CREATE TABLE EntityTags (entity_id, tag_id, properties);")
			cur.execute("CREATE TABLE Tag (id integer primary key AUTOINCREMENT, name UNIQUE);")
			con.commit()
			cur.execute("CREATE TRIGGER tag_add AFTER INSERT on Tag BEGIN select change_tag(); END;")
			cur.execute("CREATE TRIGGER tag_update AFTER UPDATE on Tag BEGIN select change_tag(); END;")
			cur.execute("CREATE TRIGGER tag_delete AFTER DELETE on Tag BEGIN select change_tag(); END;")
			con.commit()

		self.tags = Tags(con, cur)
		self.entities = Entities(con, cur, self.tags)
	
	def entities_with_tag(self, name):
		""" Returns a list of entities that have tag name """
		res = []
		for row in self.con.execute('select e.url from EntityTags et join Entity e on e.id = et.entity_id where et.tag_id=?;', (self.tags[name],)):
			res.append(row[0])
		return res
	
	def get_tag_values(self, tag):
		self.cur.execute("select distinct properties from EntityTags join Tag on tag_id == id where name = ?", (tag,))
		return [x[0] for x in self.cur.fetchall()]
